#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod

class Fetcher(object):
    def __init__(self, dataset, collate_fn, drop_last):
        self.dataset = dataset
        self.collate_fn = collate_fn
        self.drop_last = drop_last

    @abstractmethod
    def __call__(self, batched_indices):
        raise NotImplementedError

class IterableFetcher(Fetcher):
    def __init__(self, dataset, collate_fn, drop_last):
        super(IterableFetcher, self).__init__(dataset, collate_fn, drop_last)
        self.dataset_iter = iter(dataset)

    def __call__(self, batched_indices):
        data = []
        for _ in batched_indices:
            try:
                data.append(next(self.dataset_iter))
            except StopIteration:
                break
        if len(data) == 0 or (self.drop_last and len(data) < len(batched_indices)):
            raise StopIteration
        return self.collate_fn(data)

class IndexFetcher(Fetcher):
    def __init__(self, dataset, collate_fn, drop_last):
        super(IndexFetcher, self).__init__(dataset, collate_fn, drop_last)

    def __call__(self, batched_indices):
        data = [self.dataset[idx] for idx in batched_indices]
        return self.collate_fn(data)

FETCHERS = {"index": IndexFetcher, "iter": IterableFetcher, }
